/*
 * parser.c - configuration parser for lincond
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <dotconf.h>
#include <pwd.h>
#include <grp.h>
#include <math.h>
#include <stdint.h>
#include <sched.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "list.h"
#include "str.h"
#include "buffer.h"
#include "parser.h"
#include "misc.h"
#include "bo.h"

#include "main.h"
#include "parser.h"

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
					__FUNCTION__,  ##arg)

/* context flags to keep track on current context */
#define CONTEXT_ROOT            0x1
#define CONTEXT_STATION         0x2
#define CONTEXT_MODULE          0x4

/* option flags to detect duplicated options */
#define HAVE_OPTION_USER	0x01
#define HAVE_OPTION_GROUP	0x02
#define HAVE_OPTION_INPUTS	0x04
#define HAVE_OPTION_OUTPUTS	0x08
#define HAVE_OPTION_SOCKETFILE	0x10
#define HAVE_OPTION_PRIORITY	0x20
#define HAVE_OPTION_BYTEORDER	0x40

struct parser_context {
	char			*end_token;
	unsigned int		context;
	unsigned int		option_flags;
	uid_t			user;
	gid_t			group;
	char			socketfile[PATH_MAX];
	int			priority;
	int			byte_order;
	void			*inputs_bitmap, *outputs_bitmap;
	size_t			inputs_size, outputs_size;
	struct station		*current_station;
	struct module		*current_module;
	str_t			errstr;
};

extern struct slist *stations_list;
extern struct module_options module_options_table[];

static char close_station_token[] = "</station>";
static char close_module_token[] = "</module>";

static const char *opt_user(command_t *cmd, context_t *opaque);
static const char *opt_group(command_t *cmd, context_t *opaque);
static const char *opt_socketfile(command_t *cmd, context_t *opaque);
static const char *opt_inputs(command_t *cmd, context_t *opaque);
static const char *opt_outputs(command_t *cmd, context_t *opaque);
static const char *opt_usecsum(command_t *cmd, context_t *opaque);
static const char *opt_timeout(command_t *cmd, context_t *opaque);
static const char *opt_priority(command_t *cmd, context_t *opaque);
static const char *opt_byteorder(command_t *cmd, context_t *opaque);

static const char *open_station(command_t *cmd, context_t *opaque);
static const char *close_station(command_t *cmd, context_t *opaque);
static const char *opt_port(command_t *cmd, context_t *opaque);
static const char *opt_speed(command_t *cmd, context_t *opaque);
static const char *open_module(command_t *cmd, context_t *opaque);
static const char *close_module(command_t *cmd, context_t *opaque);
static const char *opt_slot(command_t *cmd, context_t *opaque);
static const char *opt_moduletype(command_t *cmd, context_t *opaque);
static const char *opt_outputsmap(command_t *cmd, context_t *opaque);
static const char *opt_inputsmap(command_t *cmd, context_t *opaque);
static const char *opt_address(command_t *cmd, context_t *opaque);
static const char *opt_units(command_t *cmd, context_t *opaque);
static const char *opt_outputchanged(command_t *cmd, context_t *opaque);

/* All options and pointer to their handlers are here. */
static configoption_t options[] = {
	{ "user", ARG_STR, opt_user,  NULL, CONTEXT_ROOT },
	{ "group", ARG_STR, opt_group, NULL, CONTEXT_ROOT },
	{ "socketfile", ARG_STR, opt_socketfile, NULL, CONTEXT_ROOT },
	{ "inputs", ARG_INT, opt_inputs, NULL, CONTEXT_ROOT },
	{ "outputs", ARG_INT, opt_outputs, NULL, CONTEXT_ROOT },
	{ "priority", ARG_INT, opt_priority, NULL, CONTEXT_ROOT },
	{ "socketfile", ARG_STR, opt_byteorder, NULL, CONTEXT_ROOT },
	{ "<station>", ARG_NONE, open_station, NULL, CONTEXT_ROOT },
	{ close_station_token, ARG_NONE, close_station, NULL, CONTEXT_STATION },
	{ "port", ARG_STR, opt_port, NULL, CONTEXT_STATION },
	{ "speed", ARG_STR, opt_speed, NULL, CONTEXT_STATION },
	{ "usecsum", ARG_TOGGLE, opt_usecsum, NULL, CONTEXT_STATION },
	{ "timeout", ARG_INT, opt_timeout, NULL, CONTEXT_STATION },
	{ "<module>", ARG_NONE, open_module, NULL, CONTEXT_STATION },
	{ close_module_token, ARG_NONE, close_module, NULL, CONTEXT_MODULE },
	{ "slot", ARG_INT, opt_slot, NULL, CONTEXT_MODULE },
	{ "moduletype", ARG_STR, opt_moduletype, NULL, CONTEXT_MODULE },
	{ "outputsmap", ARG_INT, opt_outputsmap, NULL, CONTEXT_MODULE },
	{ "inputsmap", ARG_INT, opt_inputsmap, NULL, CONTEXT_MODULE },
	{ "address", ARG_INT, opt_address, NULL, CONTEXT_MODULE },
	{ "units", ARG_STR, opt_units, NULL, CONTEXT_MODULE },
	{ "outputchanged", ARG_TOGGLE, opt_outputchanged, NULL, CONTEXT_MODULE },
	LAST_OPTION,
};

static const char *opt_user(command_t *cmd, context_t *opaque)
{
	int res;
	const char *err = NULL;
	struct passwd pwd;
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (ctx->option_flags & HAVE_OPTION_USER)
		return "duplicated option `user'";

	res = getpw_by_name_or_uid(cmd->data.str, &pwd);

	if (res) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "can't find user or uid '%.100s'", cmd->data.str);
		err = str_ptr(ctx->errstr);
	} else {
		ctx->user = pwd.pw_uid;
		ctx->option_flags |= HAVE_OPTION_USER;
	}

	return err;
}

static const char *opt_group(command_t *cmd, context_t *opaque)
{
	int res;
	const char *err = NULL;
	struct group grp;
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (ctx->option_flags & HAVE_OPTION_GROUP)
		return "duplicated option `group'";

	res = getgr_by_name_or_gid(cmd->data.str, &grp);

	if (res) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "can't find group or gid '%.100s'", cmd->data.str);
		err = str_ptr(ctx->errstr);
	} else {
		ctx->group = grp.gr_gid;
		ctx->option_flags |= HAVE_OPTION_GROUP;
	}

	return err;
}

static const char *opt_socketfile(command_t *cmd, context_t *opaque)
{
	char *canonized_filename;
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (ctx->option_flags & HAVE_OPTION_SOCKETFILE)
		return "duplicated option `socketfile'";

	if (*cmd->data.str == '/' || *cmd->data.str == '.') {
		canonized_filename = canonize_path(cmd->data.str);
		xstrncpy(ctx->socketfile, canonized_filename, sizeof(ctx->socketfile));
		xfree(canonized_filename);
	} else {
		snprintf(ctx->socketfile, sizeof(ctx->socketfile), "%s/%s/%s", LOCALSTATE_DIR, DAEMON_DIR, SOCKFILE_DEFAULT);
	}

	ctx->option_flags |= HAVE_OPTION_SOCKETFILE;
	return NULL;
}

static const char *opt_inputs(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (ctx->option_flags & HAVE_OPTION_INPUTS)
		return "duplicated option `inputs'";

	if (cmd->data.value > 0) {
		ctx->inputs_size = cmd->data.value;
		/* Each bit in this memory chunk corresponds to one byte in input image. */
		ctx->inputs_bitmap = xmalloc0(ctx->inputs_size / CHAR_BIT +  (ctx->inputs_size % CHAR_BIT ? 1 : 0));
	}

	ctx->option_flags |= HAVE_OPTION_INPUTS;

	return NULL;
}

static const char *opt_outputs(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (ctx->option_flags & HAVE_OPTION_OUTPUTS)
		return "duplicated option `outputs'";

	if (cmd->data.value > 0) {
		ctx->outputs_size = cmd->data.value;
		ctx->outputs_bitmap = xmalloc0(ctx->outputs_size / CHAR_BIT + (ctx->outputs_size % CHAR_BIT ? 1 : 0));
	}

	ctx->option_flags |= HAVE_OPTION_OUTPUTS;

	return NULL;
}

static const char *opt_priority(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (ctx->option_flags & HAVE_OPTION_PRIORITY)
		return "duplicated option `priority'";

	if (cmd->data.value > sched_get_priority_max(SCHED_RR))
		return "option 'priority' has invalid argument (too big)";
	if (cmd->data.value < sched_get_priority_min(SCHED_RR))
		return "option 'priority' has invalid argument (too small)";

	ctx->priority = cmd->data.value;
	ctx->option_flags |= HAVE_OPTION_PRIORITY;
	return NULL;
}

static const char *opt_byteorder(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *)opaque;

	if (ctx->option_flags & HAVE_OPTION_BYTEORDER)
		return "duplicated option `byteorder'";

	if (strcasecmp(cmd->data.str, "be") == 0)
		ctx->byte_order = BO_BE;
	else if (strcasecmp(cmd->data.str, "le") == 0)
		ctx->byte_order = BO_LE;
	else if (strcasecmp(cmd->data.str, "native") == 0)
		ctx->byte_order = BO_NATIVE;
	else 
		return "option `byteorder' has unrecognized argument";

	ctx->option_flags |= HAVE_OPTION_BYTEORDER;

	return NULL;
}


static const char *open_station(command_t *cmd, context_t *opaque)
{
	const char *err = NULL;
	char *old_end_token;
	struct parser_context *ctx = (struct parser_context *) opaque;
	struct station *station;

	if (ctx->context & CONTEXT_STATION)
		return "nested section <station> is not allowed";

	/* enter station context */
	old_end_token = ctx->end_token;
	ctx->end_token = close_station_token;
	ctx->context |= CONTEXT_STATION;
	/* create new station */
	station = xmalloc0(sizeof(*station));
	station->modules_list = NULL;
	station->port[0] = 0;
	station->speed[0] = 0;
	station->state = STATION_DISCONNECTED;
	station->serial_fd = -1;
	station->tout = NULL;
	station->rb = NULL;
	station->wb = NULL;
	station->has_parallel_bus = FALSE;
	station->use_csum = FALSE;
	station->response_tout = TIMEOUT_MS_DEFAULT;
	ctx->current_station = station;
	/* read options from station context */
	while (!cmd->configfile->eof) {
		err = dotconf_command_loop_until_error(cmd->configfile);
		if (err != NULL) {
			if (err == close_station_token)
				break;
			dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "%s", err);
		} else {
			err = "section <station> is not closed properly";
			break;
		}
	}

	/* If section was not properly closed it is not my problem.
	 * Station is added to the list. */

	if (strlen(station->port) > 0 && strlen(station->speed) > 0) {
		struct slist *list;
		/* Warn user if station occupies the same serial device. */
		for (list = stations_list; list != NULL; list = list->next) {
			struct station *tmp_station = (struct station *) list->data;
			if (strcmp(tmp_station->port, station->port) == 0) {
				dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "station uses serial device '%s' which is in use by another station", station->port);
				break;
			}
		}
		if (station->modules_list)
			station->modules_list = slist_reverse(station->modules_list);
		list = slist_new();
		list->data = station;
		stations_list = slist_add(stations_list, list);
	} else {
		dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "Ignoring invalid section <station> and it's contents. You must provide mandatory options 'port' and 'speed'.");
		struct slist *tmp_list;
		for (tmp_list = station->modules_list; tmp_list; tmp_list = tmp_list->next) {
			struct module *module = (struct module *) tmp_list->data;
			xfree(module);
		}
		if (station->modules_list)
			slist_free(&station->modules_list);
		/* Goodby, i don't need you... */
		xfree(station);
	}

	ctx->current_module = NULL;
	ctx->current_station = NULL;

	if (err != close_station_token)
		return err;

	/* quit station context */
	ctx->context &= ~CONTEXT_STATION;
	ctx->end_token = old_end_token;

	return NULL;
}

static const char *close_station(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	/* If option </station> is used outside station context inform user about error. */
	if ((ctx->context & CONTEXT_STATION) == 0) {
		if (ctx->end_token ) {
			str_reset(ctx->errstr);
			str_snprintf(ctx->errstr, "previous section is not closed "
				     "properly because option '%s' missing",
				     ctx->end_token);
		} else {
			str_reset(ctx->errstr);
			str_append(ctx->errstr, "missing section <station>");
		}
		return str_ptr(ctx->errstr);
	}

	/* If everything is OK return closing token pointer. */
	return close_station_token;
}

static const char *opt_port(command_t *cmd, context_t *opaque)
{
	int res;
	struct stat st;
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (strlen(ctx->current_station->port) != 0)
		return "duplicated option 'station'";

	if (cmd->data.str[0] == '/')
		xstrncpy(ctx->current_station->port, cmd->data.str, sizeof(ctx->current_station->port));
	else
		snprintf(ctx->current_station->port, sizeof(ctx->current_station->port), "/dev/%s", cmd->data.str);

	/* Check that this device exists and is really a character device */
	res = stat(ctx->current_station->port, &st);

	if (res != 0 || !S_ISCHR(st.st_mode)) {
		ctx->current_station->port[0] = 0;
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "device file '%s' does not exists or is not a character device", cmd->data.str);
		return str_ptr(ctx->errstr);
	}

	return NULL;
}

static const char *opt_speed(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (strlen(ctx->current_station->speed) != 0)
		return "duplicated option 'speed'";

	/* I don't check speed option. When the serial port
	 * is opened and speed option is wrong, function will
	 * fall back to default speed. */
	xstrncpy(ctx->current_station->speed, cmd->data.str, sizeof(ctx->current_station->speed));

	return NULL;
}

static const char *opt_usecsum(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (cmd->data.value == TRUE)
		ctx->current_station->use_csum = TRUE;
	else
		ctx->current_station->use_csum = FALSE;

	return NULL;
}

static const char *opt_timeout(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (cmd->data.value > TIMEOUT_MS_MAX) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "option 'timeout' has argument less then allowed maximum %d msec. Falling back to default.", TIMEOUT_MS_MAX);
		ctx->current_station->response_tout = TIMEOUT_MS_DEFAULT;
		return str_ptr(ctx->errstr);
	} else if (cmd->data.value < TIMEOUT_MS_MIN) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "option 'timeout' has argument more then allowed minimum %d msec. Falling back to default.", TIMEOUT_MS_MIN);
		ctx->current_station->response_tout = TIMEOUT_MS_DEFAULT;
		return str_ptr(ctx->errstr);
	} else
		ctx->current_station->response_tout = cmd->data.value;

	return NULL;
}

static inline int get_bit(void *ptr, size_t n_bit)
{
	return (((unsigned char *)(ptr))[n_bit / CHAR_BIT] & (1 << n_bit % CHAR_BIT)) > 0;
}

static inline void set_bit(void *ptr, size_t n_bit)
{
	((unsigned char *)(ptr))[n_bit / CHAR_BIT] |= (1 << n_bit % CHAR_BIT);
}

static inline void clr_bit(void *ptr, size_t n_bit)
{
	((unsigned char *)(ptr))[n_bit / CHAR_BIT] &= ~(1 << n_bit % CHAR_BIT);
}

static const char *open_module(command_t *cmd, context_t *opaque)
{
	int i;
	const char *err = NULL;
	char *old_end_token;
	struct module *module;
	struct parser_context *ctx = (struct parser_context *) opaque;

	if ((ctx->context & CONTEXT_MODULE))
		return "nested section <module> not allowed";

	/* enter module context */
	old_end_token = ctx->end_token;
	ctx->end_token = close_module_token;
	ctx->context |= CONTEXT_MODULE;

	module = xmalloc0(sizeof(*module));
	module->slot = -1;
	module->type = -1;
	module->outputs_offset = -1;
	module->inputs_offset = -1;
	module->state = MODULE_OK;
	module->unit = MODULE_UNIT_EU;
	module->output_changed = FALSE;
	module->readback_ao_chan = ~0; /* mark all AO channels require readback */
	ctx->current_module = module;

	for (i = 0; i < MAX_AO_CHANS; i++)
		module->output[i] = NAN;

	while (!cmd->configfile->eof) {
		err = dotconf_command_loop_until_error(cmd->configfile);
		if (err) {
			if (err == close_module_token)
				break;
			dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "%s", err);
		} else {
			err = "section <module> is not closed properly";
			break;
		}
	}

	if ((int) module->slot >= 0 && (int) module->type >= 0) {
		/* Check duplicated slot and mapping overlay. */
		int i, res, table_index, map_error = 0;
		struct slist *list;
		for (list = ctx->current_station->modules_list; list; list = list->next) {
			struct module *tmp_module = (struct module *) list->data;
			if (tmp_module->slot == module->slot) {
				dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "slot %d is occupied by another module", module->slot);
				/* don't forget to get rid of allocated module structure */
				xfree(module);
				break;
			}
		}
		if (list == NULL) {
			table_index = module->type;
			if (module_options_table[table_index].name_str != NULL) {
				/* Warn user if required inputs or outputs map is not set. */
				if (module_options_table[table_index].occupied_inputs &&
				    (int) module->inputs_offset < 0) {
					dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "module in slot %d requires 'inputsmap' option but it is not set", module->slot);
					module->inputs_offset = 0;
				}
				/* If this options were not set then use default value. */
				if (module_options_table[table_index].occupied_outputs &&
				    (int) module->outputs_offset < 0) {
					dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "module in slot %d requires 'outputsmap' options but it is not set", module->slot);
					module->outputs_offset = 0;
				}
				/* Now check and set overlay bits in maps */
				if (module_options_table[table_index].occupied_inputs) {
					if (module->inputs_offset + module_options_table[table_index].occupied_inputs <= ctx->inputs_size) {
						res = 0;
						for (i = 0; i < module_options_table[table_index].occupied_inputs; i++)
							res += get_bit(ctx->inputs_bitmap, module->inputs_offset + i);
						if (res) {
							map_error = TRUE;
							dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "inputs mapping for module in slot %d overlays another module", module->slot);
						}
					} else {
						map_error = TRUE;
						dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "inputs for module in slot %d map outside inputs image", module->slot);
					}
				}
				if (module_options_table[table_index].occupied_outputs) {
					if (module->outputs_offset + module_options_table[table_index].occupied_outputs <= ctx->outputs_size) {
						res = 0;
						for (i = 0; i < module_options_table[table_index].occupied_outputs; i++)
							res += get_bit(ctx->outputs_bitmap, module->outputs_offset + i);
						if (res) {
							map_error = TRUE;
							dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "outputs mapping for module in slot %d overlays another module", module->slot);
						}
					} else {
						map_error = TRUE;
						dotconf_warning(cmd->configfile, DCLOG_WARNING, 0, "outputs for module in slot %d map outside outputs image", module->slot);
					}
				}
				if (!map_error) {
					/* set bits in inputs, outputs or both bitmaps */
					if (module_options_table[table_index].occupied_outputs &&
					    module->outputs_offset + module_options_table[table_index].occupied_outputs <= ctx->outputs_size) {
						for (i = 0; i < module_options_table[table_index].occupied_outputs; i++)
							set_bit(ctx->outputs_bitmap, module->outputs_offset + i);
					}
					if (module_options_table[table_index].occupied_inputs &&
					    module->inputs_offset + module_options_table[table_index].occupied_inputs <= ctx->inputs_size) {
						for (i = 0; i < module_options_table[table_index].occupied_inputs; i++)
							set_bit(ctx->inputs_bitmap, module->inputs_offset + i);
					}
					list = slist_new();
					list->data = module;
					ctx->current_station->modules_list = slist_add(ctx->current_station->modules_list, list);
				} else
					xfree(module);
			}
		}
	}

	if (err != close_module_token)
		return err;

	ctx->current_module = NULL;
	/* quit module context */
	ctx->end_token = old_end_token;
	ctx->context &= ~CONTEXT_MODULE;
	return NULL;
}

static const char *close_module(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	/* If option </module> is used outside module context warn user. */
	if ((ctx->context & CONTEXT_MODULE) == 0) {
		if (ctx->end_token ) {
			str_reset(ctx->errstr);
			str_snprintf(ctx->errstr, "previous section is not closed "
				     "properly because option '%s' missing",
				     ctx->end_token);
		} else {
			str_reset(ctx->errstr);
			str_append(ctx->errstr, "missing section <module>");
		}
		return str_ptr(ctx->errstr);
	}

	/* If everything is OK return closing token pointer. */
	return close_module_token;
}

static const char *opt_address(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (cmd->data.value < DCON_ADDRESS_MIN) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Option 'address' is less then allowed minimum value %d.", DCON_ADDRESS_MIN);
		return str_ptr(ctx->errstr);
	}

	if (cmd->data.value > DCON_ADDRESS_MAX) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Option 'address' is more then allowed maximum value %d.", DCON_ADDRESS_MAX);
		return str_ptr(ctx->errstr);
	}

	ctx->current_module->address = cmd->data.value;
	return NULL;
}

static const char *opt_slot(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

#ifdef _USE_LIBI8K
	if (strlen(ctx->current_station->port)) {
		if (strcmp(ctx->current_station->port, "/dev/ttySA2") == 0) {
			/* We are local lincon station. */
			if (cmd->data.value < LINCON_SLOT_MIN) {
				str_reset(ctx->errstr);
				str_snprintf(ctx->errstr, "option 'slot' is less then minimum Lincon slot number %d allowed", LINCON_SLOT_MIN);
				return str_ptr(ctx->errstr);
			}
			if (cmd->data.value > LINCON_SLOT_MAX) {
				str_reset(ctx->errstr);
				str_snprintf(ctx->errstr, "option 'slot' is more then maximum Lincon slot number %d allowed", LINCON_SLOT_MAX);
				return str_ptr(ctx->errstr);
			}
			ctx->current_module->slot = cmd->data.value;
		} else {
			/* We are remote station */
			if (cmd->data.value < REMOTE_SLOT_MIN) {
				str_reset(ctx->errstr);
				str_snprintf(ctx->errstr, "option `slot' is less then minimum I/O expansion slot number %d allowed.", REMOTE_SLOT_MIN);
				return str_ptr(ctx->errstr);
			}
			if (cmd->data.value > REMOTE_SLOT_MAX) {
				str_reset(ctx->errstr);
				str_snprintf(ctx->errstr, "option `slot' is more then maximum I/O expansion slot number %d allowed.", REMOTE_SLOT_MAX);
				return str_ptr(ctx->errstr);
			}
			ctx->current_module->slot = cmd->data.value;
		}
	} else {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "You must define station's `port' option before <module>. Otherwise I can not determine station type.");
		return str_ptr(ctx->errstr);
	}
#else
	if (cmd->data.value < REMOTE_SLOT_MIN) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "option `slot' is less than minimum slot number %d allowed", REMOTE_SLOT_MIN);
		return str_ptr(ctx->errstr);
	}

	if (cmd->data.value > REMOTE_SLOT_MAX) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "option `slot' is more than maximum slot number %d allowed", REMOTE_SLOT_MAX);
		return str_ptr(ctx->errstr);
	}

	ctx->current_module->slot = cmd->data.value;
#endif

	return NULL;
}

static const char *opt_moduletype(command_t *cmd, context_t *opaque)
{
	int i;
	char *err = NULL;
	struct parser_context *ctx = (struct parser_context *) opaque;

	for (i = 0; module_options_table[i].name_str != NULL; i++) {
		if (strcmp(module_options_table[i].name_str, cmd->data.str) == 0) {
			ctx->current_module->type = module_options_table[i].module_type;
			break;
		}
	}

	/* What if module type was not identified?? */
	if (module_options_table[i].name_str == NULL) {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Module type '%.100s' given as argument to option `moduletype' is not known to me.", cmd->data.str);
		err = str_ptr(ctx->errstr);
	}

	return err;
}

static const char *opt_units(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (strcmp(cmd->data.str, "eu") == 0)
		ctx->current_module->unit = MODULE_UNIT_EU;
	else if (strcmp(cmd->data.str, "raw") == 0)
		ctx->current_module->unit = MODULE_UNIT_RAW;
	else if (strcmp(cmd->data.str, "ohm") == 0)
		ctx->current_module->unit = MODULE_UNIT_OHM;
	else if (strcmp(cmd->data.str, "fsr") == 0)
		ctx->current_module->unit = MODULE_UNIT_FSR;
	else {
		str_reset(ctx->errstr);
		str_snprintf(ctx->errstr, "Invalid argument '%.100s' to option `units'. I will use default units for this module.", cmd->data.str);
		return str_ptr(ctx->errstr);
	}
	return NULL;
}

static const char *opt_outputchanged(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	if (cmd->data.value == TRUE)
		ctx->current_module->output_changed = TRUE;
	else
		ctx->current_module->output_changed = FALSE;
	return NULL;
}

static const char *opt_outputsmap(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	ctx->current_module->outputs_offset = cmd->data.value;

	return NULL;
}

static const char *opt_inputsmap(command_t *cmd, context_t *opaque)
{
	struct parser_context *ctx = (struct parser_context *) opaque;

	ctx->current_module->inputs_offset = cmd->data.value;

	return NULL;
}

static int error_handler(configfile_t *config, int type, long dc_errno, const char *msg)
{
	warn("parser: %s: file %s line %ld", msg, config->filename, config->line);

	return 0;
}

static const char *context_checker(command_t *cmd, unsigned long context_flags)
{
	struct parser_context *ctx = (struct parser_context *) cmd->context;

	if ((context_flags & CONTEXT_ROOT) && ctx->context == 0)
		return NULL;

	if ((context_flags & ctx->context) == 0)
		return "option not allowed in this context";

	return NULL;
}
/*
static void parser_show_configuration()
{
	struct slist *station_tmp, *module_tmp;

	for (station_tmp = stations_list; station_tmp != NULL; station_tmp = station_tmp->next) {
		struct station *station = (struct station *) station_tmp->data;
		printf(" Station Configuration \n"
		       "      port: %s \n"
		       "     speed: %s \n"
		       "      csum: %s \n"
		       "   timeout: %d \n",
		       station->port,
		       station->speed,
		       station->use_csum ? "y" : "n",
		       station->response_tout);
		for (module_tmp = station->modules_list; module_tmp != NULL; module_tmp = module_tmp->next) {
			struct module *module = (struct module *) module_tmp->data;
			printf(" Module Configuration \n"
			       "      slot: %d \n"
			       "   address: %d \n"
			       "      type: %d \n"
			       "  inputmap: %ld \n"
			       " outputmap: %ld \n",
			       module->slot,
			       module->address,
			       module->type,
			       (unsigned long) module->inputs_offset,
			       (unsigned long) module->outputs_offset);
		}
	}
}
*/
int parse_configuration_file(const char *configfile, struct configfile_options *opts)
{
	const char *err = NULL;
	configfile_t *config;
	struct parser_context *ctx;

	ctx = xmalloc0(sizeof(*ctx));
	ctx->end_token = NULL;
	ctx->context = 0;
	ctx->errstr = str_new();
	ctx->current_module = NULL;
	ctx->current_station = NULL;
	ctx->group = -1;
	ctx->user = -1;
	ctx->priority = -1;
	ctx->byte_order = -1;
	ctx->inputs_size = 0;
	ctx->outputs_size = 0;
	ctx->option_flags = 0;
	/* open configuration file */
	config = dotconf_create((char *)configfile, options, (context_t *) ctx, DONT_SUBSTITUTE);
	if (!config) {
		warn("dotconf can't read file '%s'", configfile);
		str_free(&ctx->errstr);
		xfree(ctx);
		return -1;
	}
	/* set error handler and context checker */
	config->errorhandler = (dotconf_errorhandler_t) error_handler;
	config->contextchecker = context_checker;
	/* parse configuration file and warn on errors */
	while (!config->eof) {
		err = dotconf_command_loop_until_error(config);
		if (err) {
			dotconf_warning(config, DCLOG_WARNING, 0, "%s", err);
		} else
			break;
	}
	/* copy options to resulting structure */
	opts->inputs_size = ctx->inputs_size;
	opts->outputs_size = ctx->outputs_size;
	opts->group = ctx->group;
	opts->user = ctx->user;
	opts->priority = ctx->priority;
	opts->byte_order = ctx->byte_order;
	xstrncpy(opts->socketfile, ctx->socketfile, sizeof(opts->socketfile));
	/* cleanup dotconf resources */
	dotconf_cleanup(config);
	/* free resources */
	if (ctx->inputs_bitmap)
		xfree(ctx->inputs_bitmap);
	if (ctx->outputs_bitmap)
		xfree(ctx->outputs_bitmap);
	str_free(&ctx->errstr);
	xfree(ctx);

	if (stations_list)
		stations_list = slist_reverse(stations_list);

	/*parser_show_configuration();*/

	return 0;
}
